/*-
 * Copyright (C) 2007 Erik Larsson
 * 
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

import org.catacombae.rarx.*;
import java.io.*;
import java.nio.ByteBuffer;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Calendar;
import java.util.Date;

/**
 * The FuseModule class is the key class in the Java side of rarfs. Most if not all of the fuse operations are
 * redirected to static methods of this class.
 */
public class FuseModule {
    /** Constant variable that determines if debug-messages should be printed to System.err. */
    private static final boolean DEBUG = false;
    
    /** The file object corresponding to the RAR file that this filesystem represents. */
    private static File rarFile;

    // /** A list of all the entries inside the RAR file. Initialized in the init method, and never modified. */
    // private static RARFileEntry[] entries = null;

    /** A tree representing the file system hierarchy. Initialized in the init method, and never modified. */
    private static Tree<RARFileEntry> fsTree = null;

    /** Stores all the open streams corresponding to open files. */
    private static Hashtable<String, RARFileEntryStream> openFiles = new Hashtable<String, RARFileEntryStream>();
    
    /** Only used by <code>rarfsReaddir</code> as a temporary storage list.  */
    private static final LinkedList<String> readdirResult = new LinkedList<String>();
    
    /** Buffer used while transfering data from the RAR file to the output buffer. */
    private static final byte[] tempBuffer = new byte[4096];
    
    /** This method must always be called before calling any of the other methods of FuseModule. It opens the
	file handle to the RAR file and caches the file list in memory (modifications in the underlying RAR is
	_not_ recommended while it's mounted, but that should be obvious).
	@param filename the full path to the RAR file 
	@return 0 if all is well, -1 if an IOException occurred while listing the files (more error codes
	might be added in the future) */
    public static int init(String filename) {
	try {
	    debug_log("init(" + filename + ");");
	    rarFile = new File(filename);
	    //entries = ListFilesInArchive.listFiles(rarFile);
	    debug_log("SETTING fsTree HERE!!!");
	    fsTree = ListFilesInArchive.listFilesAsTree(rarFile);
	    debug_log("fsTree is INITITALIZED to " + fsTree);
	    return 0;
	} catch(Exception e) {
	    e.printStackTrace();
	    return -1;
	}
    }
    
    /** 
     * Equivalent to the read operation in struct fuse_operations.
     * @param path the path inside the file system to the file from which we want to read (starting with "/")
     * @param buf the buffer where the data is to be stored
     * @param offsetInFile the offset in the file where we want to start reading
     * @param len the number of bytes to read
     * @return a positive number indicating the number of bytes that have been read, or a negative number if
     *         something went wrong. -1 means the file has not yet been opened, -2 means an exception occurred
     *         while reading data from the file, 0 means no bytes could be read (EOF).
     */
    public static int rarfsRead(String path, byte[] buf, long offsetInFile, int len) {
	try {
	    debug_log("rarfsRead{1}(" + path + ", " + buf + ", " + offsetInFile + ", " + len + ");");
	    RARFileEntryStream stream = openFiles.get(path);
	    if(stream == null) {
		TreeNode<RARFileEntry> entry = getEntry(path);
		if(entry != null) {
		    if(entry.getValue().getHeader(0).getMethod() != NewFileHeader.METHOD_UNCOMPRESSED) {
			return 0; // If we try to read from a compressed file, it will appear as empty.
		    }
		}
		return -1;
	    }
	    else {
		stream.seek(offsetInFile);
		return stream.read(buf, 0, len);
	    }
	} catch(Exception e) { e.printStackTrace(); return -2; }
    }

    /** 
     * Equivalent to the read operation in struct fuse_operations.
     * @param path the path inside the file system to the file from which we want to read (starting with "/")
     * @param buf the buffer where the data is to be stored
     * @param offsetInFile the offset in the file where we want to start reading
     * @param len the number of bytes to read
     * @return a positive number indicating the number of bytes that have been read, or a negative number if
     *         something went wrong. -1 means the file has not yet been opened, -2 means an exception occurred
     *         while reading data from the file, -3 if buf is read only, -4 if buf does not have enough capacity,
     *         0 means no bytes could be read (EOF).
     */
    public static int rarfsRead(String path, ByteBuffer buf, long offsetInFile, int len) {
	try {
	    debug_log("rarfsRead{2}(" + path + ", " + buf + ", " + offsetInFile + ", " + len + ");");
	    RARFileEntryStream stream = openFiles.get(path);
	    if(stream == null) {
		TreeNode<RARFileEntry> entry = getEntry(path);
		if(entry != null) {
		    if(entry.getValue().getHeader(0).getMethod() != NewFileHeader.METHOD_UNCOMPRESSED) {
			return 0; // If we try to read from a compressed file, it will appear as empty.
		    }
		}
		return -1;
	    }
	    else {
		debug_log("buf.isReadOnly(): " + buf.isReadOnly());
		debug_log("buf.hasArray(): " + buf.hasArray());
		debug_log("buf.isDirect(): " + buf.isDirect());
		if(buf.isReadOnly())
		    return -3;
		else if(buf.capacity() < len)
		    return -4;

		stream.seek(offsetInFile);
		int bytesRead = 0;
		while(bytesRead < len) {
		    int curBytesRead;
		    if(buf.hasArray()) {
			byte[] backingArray = buf.array();
			curBytesRead = stream.read(backingArray, 0, len);
			if(curBytesRead == -1)
			    break;
		    }
		    else {
			curBytesRead = stream.read(tempBuffer, 0, (len < tempBuffer.length?len:tempBuffer.length));
			if(curBytesRead == -1)
			    break;
			buf.put(tempBuffer, 0, curBytesRead);
		    }
		    
		    bytesRead += curBytesRead;
		}
		return bytesRead;
	    }
	} catch(Exception e) { e.printStackTrace(); return -2; }
    }    

    /** 
     * Equivalent to the open operation in struct fuse_operations.
     * @param path the (absolute) pathname inside this filesystem to the file that is to be opened (starting with "/")
     * @return 0 if all went well, -1 if the file could not be found, -2 if an exception occurred while trying to
     *         open a stream to the entry in the RAR file corresponding to the file system entry
     */
    public static int rarfsOpen(String path) {
	try {
	    debug_log("rarfsOpen(" + path + ");");
	    TreeNode<RARFileEntry> node = getEntry(path);
	    if(node != null) {
		RARFileEntry entry = node.getValue();
		if(entry != null) {
		    if(entry.getHeader(0).getMethod() == NewFileHeader.METHOD_UNCOMPRESSED) {
			openFiles.put(path, new RARFileEntryStream(node.getValue()));
			return 0;
		    }
		    else
			return 0; // We just pretend opening the file goes well.
		}
		else
		    return -1;
	    }
	    else
		return -1;
	} catch(Exception e) { e.printStackTrace(); return -2; }

    }

    /**
     * Equivalent to the readdir operation in struct fuse_operations.
     * Currently this does not correctly deal with directory entries inside rar files.
     * @param path the (absolute) pathname inside this filesystem to the dir that is to be read (starting with "/")
     * @return an array of filename strings
     */
    public static String[] rarfsReaddir(String path) {
	try {
	    debug_log("rarfsReaddir(" + path + ");");
	    LinkedList<String> result = readdirResult;
	    result.clear();
	    TreeNode<RARFileEntry> dirEntry = getEntry(path);
	    System.out.println("got: " + dirEntry);
	    if(dirEntry != null && dirEntry instanceof Tree) {
		Tree<RARFileEntry> t = (Tree<RARFileEntry>)dirEntry;
		RARFileEntry[] entries = t.listChildren(new RARFileEntry[t.childCount()]);
		String[] names = t.listChildrenNames();
		debug_log("names.length="+names.length+" entries.length="+entries.length);
		for(int i = 0; i < entries.length; ++i) {
		    debug_log("Checking \"" + names[i] + "\"");
		    //if(entries[i].getHeader(0).getMethod() == RARHeader.METHOD_UNCOMPRESSED)
		    // Note: changed to allow all files to be listed
		    result.addLast("/" + names[i]);
		}
		return result.toArray(new String[result.size()]);
	    }
	} catch(Exception e) {
	    e.printStackTrace();
	}
	return null;
    }

    /**
     * Equivalent to the getattr operation in struct fuse_operations.
     * @param path the (absolute) pathname inside this filesystem to the requested file or folder
     * @param res the Stat object where we put the result of the getattr operation (output parameter)
     * @return 0 if operation completed without problems, -1 if an exception occurred (more info is printed to System.err),
     *         -2 if the pathname doesn't exist in the file system, -3 if there is no entry in the RAR file for the pathname,
     *         -4 if <code>res</code> is <code>null</code>
     */
    public static int rarfsGetattr(String path, Stat res) {
	try {
	    debug_log("rarfsGetattr(" + path + ", " + res + ");");
	    if(res == null)
		return -4;
	    
	    TreeNode<RARFileEntry> node = getEntry(path);
	    if(node != null) {
		RARFileEntry entry = node.getValue();
		//Stat res = new Stat();
		if(entry != null) {
		    res.st_nlink = 1;
		    Date lastModified = entry.getHeader(0).getFileTimeAsDate();
		    Calendar c = Calendar.getInstance();
		    c.setTime(lastModified);
		    res.setAccessTime(c);
		    res.setModifyTime(c);
		    res.setCreateTime(c);
		    
		    if(node instanceof Tree) {
			res.st_mode = (short)(Stat.S_IFDIR | 0444); // Read permission for all, directory
			res.st_size = 0;
		    }
		    else {
			res.st_mode = (short)(Stat.S_IFREG | 0444); // Read permission for all, standard file
			
			if(entry.getHeader(0).getMethod() == NewFileHeader.METHOD_UNCOMPRESSED)
			    res.st_size = entry.getHeader(0).getUnpSize();
			else
			    res.st_size = 0;
		    }
		    return 0;
		}
		else
		    return -3;
	    }
	    else
		return -2;
	} catch(Exception e) { e.printStackTrace(); return -1; }
    }

    /**
     * Parses the pathname <code>path</code> and looks up the corresponding TreeNode.
     * @param path the absolute pathname to the requested file system entry (must begin with a "/")
     * @return the corresponding TreeNode, or <code>null</code> if it can't be found
     */
    private static TreeNode<RARFileEntry> getEntry(String path) {
	debug_log("getEntry(" + path + ")");
	String truePath = path.substring(1);
	if(truePath.endsWith("/"))
	    truePath = truePath.substring(0, truePath.length()-1);
	debug_log("truePath=" + truePath);
	if(truePath.trim().equals("")) {
	    if(fsTree == null)
		debug_log("fsTree == null!!!");
	    return fsTree;
	}

	String[] pathComponents = truePath.split("/"); // THIS IS A FUCKING UNIX SYSTEM, NO FREAKIN SLASHES ARE ALLOWED IN THE FILENAMES INSIDE THE RARS YOU MOUNT HERE. TRY ANYWAY AND YOU DIE!! (really, RAR won't make archives with such filenames, I think..)
	debug_log("pathComponents.length==" + pathComponents.length);

	Tree<RARFileEntry> currentTree = fsTree;
	for(int i = 0; i < pathComponents.length-1; ++i) {
	    TreeNode<RARFileEntry> child = currentTree.get(pathComponents[i]);
	    if(child instanceof Tree)
		currentTree = (Tree<RARFileEntry>)child;
	    else
		return null;
	}
	
	TreeNode<RARFileEntry> requestedNode = currentTree.get(pathComponents[pathComponents.length-1]);
	return requestedNode;


// 	for(int i = 0; i < entries.length; ++i)
// 	    if(entries[i].getHeader(0).getFilenameAsString().equals(truePath))
// 		return entries[i];
// 	return null;
    }

    private static void debug_log(String msg) {
	if(DEBUG)
	    System.err.println(msg);
    }
}