/*-
 * Copyright (C) 2007 Erik Larsson
 * 
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import java.util.*;
import java.io.*;
import com.apple.eawt.*;

public class MountListWindow extends JFrame implements ApplicationListener {
    private JList mountList;
    private JScrollPane mountListScroller;
    private JButton mountNewButton;
    private JButton unmountButton;
    private Vector<Mount> mounts = new Vector<Mount>();
    private final JFileChooser fileChooser = new JFileChooser();
    
    public MountListWindow() {
	super("List of mounted RAR files");
	//setLayout(new BorderLayout(5, 5));

	mountList = new JList();
	mountListScroller = new JScrollPane(mountList);
	mountNewButton = new JButton("Mount RAR file");
	unmountButton = new JButton("Unmount selected");

	JPanel mountListScrollerWrapper = new JPanel();
	mountListScrollerWrapper.setLayout(new BorderLayout());
	mountListScrollerWrapper.add(mountListScroller, BorderLayout.CENTER);
	//mountListScrollerWrapper.setOpaque(false);
	
	JPanel temp = new JPanel();
	temp.add(mountNewButton);
	temp.add(unmountButton);
	
	mountListScrollerWrapper.setBorder(new EmptyBorder(5, 5, 0, 5));
	temp.setBorder(new EmptyBorder(0, 5, 5, 5));
	
	
	add(mountListScrollerWrapper, BorderLayout.CENTER);
	add(temp, BorderLayout.SOUTH);

	unmountButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    int[] selection = mountList.getSelectedIndices();
		    for(int i : selection)
			unmountAction(i);
		}
	    });
	mountNewButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    fileChooser.setMultiSelectionEnabled(false);
		    fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
		    if(fileChooser.showOpenDialog(MountListWindow.this) == JFileChooser.APPROVE_OPTION) {
			File f = fileChooser.getSelectedFile();
			mountRar(f);
		    }
		}
	    });
	addWindowListener(new WindowAdapter() {
		public void windowClosing(WindowEvent e) {
		    exitApplication();
		}
	    });

	pack();
	setLocationRelativeTo(null);
    }
    
    public void addMount(Mount m) {
	mounts.add(m);
	mountList.setListData(mounts);
    }

    public void removeMount(int index) {
	mounts.remove(index);
	mountList.setListData(mounts);
    }
    
    public void unmountAction(int index) {
	Mount m = mounts.get(index);
	try {
	    if(m.mountDir.exists()) {
		int unmountCode = Launcher.unmount(m, false);
		if(unmountCode == -2)
		    JOptionPane.showMessageDialog(this, "Could not unmount. Check if you have any open files in the mount.",
						  "Error", JOptionPane.ERROR_MESSAGE);
		else if(unmountCode == -1 && m.mountDir.exists()) {
		    JOptionPane.showMessageDialog(this, "Volume unmounted, but the mount folder " + m.mountDir +
						  " could not be removed. Remember to remove it manually.",
						  "Unmount partially complete", JOptionPane.WARNING_MESSAGE);
		    removeMount(index);
		}
		else {
		    JOptionPane.showMessageDialog(this, "Volume unmounted.", "Information", JOptionPane.INFORMATION_MESSAGE);
		    removeMount(index);
		}
	    }
	    else {
		JOptionPane.showMessageDialog(this, "Volume was already unmounted.", "Information", JOptionPane.INFORMATION_MESSAGE);
		removeMount(index);
	    }
	}
	catch(Exception e) {
	    StackTraceElement[] elems = e.getStackTrace();
	    String message = "umount threw exception(" + e + ")\n";
	    for(StackTraceElement elem : elems)
		message += elem.toString() + "\n";
	    JOptionPane.showMessageDialog(this, message, "Error", JOptionPane.ERROR_MESSAGE);
	}
    }

    private void exitApplication() {
	if(mounts.size() == 0)
	    System.exit(0);

	Application.getApplication().removeApplicationListener(this);

	boolean mountExecuted = false;
	LinkedList<String[]> resList = new LinkedList<String[]>();
	try {
	    Process p = Runtime.getRuntime().exec("mount");
	    BufferedReader is = new BufferedReader(new InputStreamReader(p.getInputStream()));
	    for(String currentLine = is.readLine(); currentLine != null; currentLine = is.readLine())
		resList.add(currentLine.split("\\s"));
	    p.waitFor();
	    mountExecuted = p.exitValue() == 0;
	    if(!mountExecuted)
		JOptionPane.showMessageDialog(this, "mount exit code " + p.exitValue(), "debug", JOptionPane.PLAIN_MESSAGE);
	}
	catch(Exception e) {
	    StackTraceElement[] elems = e.getStackTrace();
	    String message = "mount threw exception(" + e + ")\n";
	    for(StackTraceElement elem : elems)
		message += elem.toString() + "\n";
	    JOptionPane.showMessageDialog(this, message, "Exception", JOptionPane.ERROR_MESSAGE);
	}
	    
	LinkedList<Mount> activeMounts = new LinkedList<Mount>();
	for(Mount m : mounts) {
	    if(m.mountDir.exists() && m.mountDir.isDirectory()) {
		if(!mountExecuted)
		    activeMounts.addLast(m); // We can't know for sure if it's really mounted, so let's just say it is
		else {
		    resCheck:
		    for(String[] currentLine : resList) {
			for(String currentWord : currentLine) {
			    if(currentWord.equals(m.mountDir.getAbsolutePath())) {
				activeMounts.addLast(m);
				break resCheck;
			    }
			}
		    }
		}
	    }
	}
	
	if(activeMounts.size() > 0) {
	    String mountList = "";
	    for(Mount m : activeMounts)
		mountList += "    " + m.mountDir.getAbsolutePath() + "\n";
	    int res = JOptionPane.showConfirmDialog(this, "You have " + activeMounts.size() + " active mounts:\n" +
						    mountList +
						    "Do you want to unmount them before exiting?\n" +
						    "(If you choose not to, you will have to manually unmount them later)", 
						    activeMounts.size() + " mounts still active",
						    JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
	    if(res == JOptionPane.YES_OPTION) {
		for(Mount m : activeMounts) {
		    int unmountCode = 0;
		    try { unmountCode = Launcher.unmount(m, false); }
		    catch(Exception e) {
			StackTraceElement[] elems = e.getStackTrace();
			String message = "umount threw exception(" + e + ")\n";
			for(StackTraceElement elem : elems)
			    message += elem.toString() + "\n";
			JOptionPane.showMessageDialog(this, message, "Error", JOptionPane.ERROR_MESSAGE);
			continue;
		    }
		    if(unmountCode != 0) {
			if(unmountCode == -2) {
			    res = JOptionPane.showConfirmDialog(this, 
								"Could not unmount " + m.mountDir + 
								". Do you want to try to force unmount?\n" +
								"(any open handles to the mounted folder will be invalidated)", 
								"Cannot unmount",
								JOptionPane.YES_NO_OPTION, JOptionPane.ERROR_MESSAGE);
			    if(res == JOptionPane.YES_OPTION) {
				int secondUnmountAttempt = 0;
				try { secondUnmountAttempt = Launcher.unmount(m, true); }
				catch(Exception e) {
				    StackTraceElement[] elems = e.getStackTrace();
				    String message = "umount threw exception(" + e + ")\n";
				    for(StackTraceElement elem : elems)
					message += elem.toString() + "\n";
				    JOptionPane.showMessageDialog(this, message, "Error", JOptionPane.ERROR_MESSAGE);
				    continue;
				}

				if(secondUnmountAttempt == -2)
				    JOptionPane.showMessageDialog(this, "Still can't unmount. You will have to unmount manually.",
								  "Error", JOptionPane.ERROR_MESSAGE);
				else if(secondUnmountAttempt == -1)
				    JOptionPane.showMessageDialog(this, "Volume unmounted, but the mount folder " + m.mountDir +
								  " could not be removed. Remember to remove it manually.",
								  "Unmount partially complete", JOptionPane.WARNING_MESSAGE);
			    }
			}
			else if(unmountCode == -1)
			    JOptionPane.showMessageDialog(this, "Volume unmounted, but the mount folder " + m.mountDir +
							  " could not be removed. Remember to remove it manually.",
							  "Unmount partially complete", JOptionPane.WARNING_MESSAGE);
		    }
		}
		System.exit(0);
	    }
	    else if(res == JOptionPane.NO_OPTION) {
		System.exit(0);
	    }
	    else if(res == JOptionPane.CANCEL_OPTION) {
		Application.getApplication().addApplicationListener(this);
	    }
	}
	else
	    System.exit(0);
    }

    private void mountRar(File rarFile) {
	try {
	    File[] mountDirWrapper = new File[1];
	    int exitCode = Launcher.mount(rarFile, mountDirWrapper);
	    File mountDir = mountDirWrapper[0];
	    if(exitCode == -1)
		JOptionPane.showMessageDialog(this, "An attempt to mount " + rarFile.getAbsolutePath() + " failed.", "Error", JOptionPane.ERROR_MESSAGE);
	    else if(exitCode >= 0) {
		addMount(new Mount(rarFile, mountDir));
		if(exitCode == 2)
		    JOptionPane.showMessageDialog(this, "Mounted " + rarFile.getName() + " at " + mountDir.getAbsolutePath() + ".", "Mount successful", JOptionPane.INFORMATION_MESSAGE);
	    }
	    else
		JOptionPane.showMessageDialog(this, "Unknown error (" + exitCode + ") when trying to mount " + rarFile.getName() + " at " + mountDir.getAbsolutePath() + ".", "Error", JOptionPane.ERROR_MESSAGE);
	}
	catch(Exception e) {
	    StackTraceElement[] elems = e.getStackTrace();
	    String message = "mount threw exception(" + e + ")\n";
	    for(StackTraceElement elem : elems)
		message += elem.toString() + "\n";
	    JOptionPane.showMessageDialog(this, message, "Error", JOptionPane.ERROR_MESSAGE);
	}
    }

    // ApplicationListener implementation below
    
    public void handleAbout(ApplicationEvent event) {
	//System.err.println("handleAbout event received (filename = " + event.getFilename() + ")");
	JOptionPane.showMessageDialog(null, "RARMounter 0.1\nCopyright 2007 Erik Larsson\nLicensed under the GNU General Public License", "About", JOptionPane.INFORMATION_MESSAGE);
	event.setHandled(true);
    }
    public void handleOpenApplication(ApplicationEvent event) {
	//System.err.println("handleOpenApplication event received (filename = " + event.getFilename() + ")");
    }
    public void handleOpenFile(ApplicationEvent event) {
	//JOptionPane.showMessageDialog(null, "handleOpenFile event received (filename = " + event.getFilename() + ")", "Info", JOptionPane.INFORMATION_MESSAGE);
	mountRar(new File(event.getFilename()));
	event.setHandled(true);
    }
    public void handlePreferences(ApplicationEvent event) {
	//System.err.println("handlePreferences event received (filename = " + event.getFilename() + ")");
    }
    public void handlePrintFile(ApplicationEvent event) {
	//System.err.println("handlePrintFile event received (filename = " + event.getFilename() + ")");
    }
    public void handleQuit(ApplicationEvent event) {
	//System.err.println("handleQuit event received (filename = " + event.getFilename() + ")");
	exitApplication();
    }
    public void handleReOpenApplication(ApplicationEvent event) {
	//System.err.println("handleReOpenApplication event received (filename = " + event.getFilename() + ")");
    }
}