/*-
 * Copyright (C) 2007 Erik Larsson
 * 
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

#include "rarfs_fuse.h"
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <jni.h>

/* For getcwd... */
#include <unistd.h>
#include <sys/param.h>

#define FALSE 0
#define TRUE 1
#define USER_CLASSPATH "./java-src:./lib/rarx.jar"
#define HELLO_ENABLED FALSE /* true: the old hello world filesystem is concatenated into the fs tree */
#define USEDIRECTBUFFER TRUE /* true: use java.nio.ByteBuffer to directly read into native-code memory locations */
#define DEBUGMODE FALSE
#if DEBUGMODE
#define DEBUG(...) fprintf(debugFile, __VA_ARGS__); fflush(debugFile);
#else
#define DEBUG(...) ;
#endif

static const char *hello_str = "Hello World!\n";
static const char *hello_path = "/hello";
static FILE *debugFile = NULL;
static JavaVM *staticJvm = NULL;
static char *rarFilename = NULL;
static char workingDirectory[MAXPATHLEN]; // <- a weakness. we should find a way to get cwd without the MAXPATHLEN limitation


static void printProperty(const char* propertyName, JNIEnv *env, jclass cls, jmethodID mid) {
  jstring userdotdir = (*env)->NewStringUTF(env, propertyName);
  jstring userdir = (*env)->CallStaticObjectMethod(env, cls, mid, userdotdir);
  const char *chars = (*env)->GetStringUTFChars(env, userdir, NULL);
  DEBUG("  - %s = \"%s\"\n", propertyName, (char*)chars);
  (*env)->ReleaseStringUTFChars(env, userdir, chars);
}

static void printJVMInfo(JavaVM *jvm, JNIEnv *env) {
/*   FILE *testFile = fopen("./printjvminfo_created_this_file", "w"); */
/*   fprintf(testFile, "hej"); */
/*   fclose(testFile); */

  DEBUG("  - Find class java.lang.System...\n");
  jclass cls = (*env)->FindClass(env, "java/lang/System");
  DEBUG("  - Get static method getProperty...\n");
  jmethodID mid = (*env)->GetStaticMethodID(env, cls, "getProperty", "(Ljava/lang/String;)Ljava/lang/String;");
  
  printProperty("user.name", env, cls, mid);
  printProperty("user.home", env, cls, mid);
  printProperty("user.dir", env, cls, mid);
  printProperty("java.class.path", env, cls, mid);

}

static int createJavaVM() {
  JNIEnv *env;
  JavaVM *jvm;
  jint res;
  
#ifdef JNI_VERSION_1_2
  JavaVMInitArgs vm_args;
  JavaVMOption options[1];
  options[0].optionString =
    "-Djava.class.path=" USER_CLASSPATH;
  vm_args.version = 0x00010002;
  vm_args.options = options;
  vm_args.nOptions = 1;
  vm_args.ignoreUnrecognized = JNI_TRUE;
  /* Create the Java VM */
  res = JNI_CreateJavaVM(&jvm, (void**)&env, &vm_args);
#else
  JDK1_1InitArgs vm_args;
  char classpath[1024];
  vm_args.version = 0x00010001;
  JNI_GetDefaultJavaVMInitArgs(&vm_args);
  /* Append USER_CLASSPATH to the default system class path */
  sprintf(classpath, "%s%c%s",
	  vm_args.classpath, PATH_SEPARATOR, USER_CLASSPATH);
  vm_args.classpath = classpath;
  /* Create the Java VM */
  res = JNI_CreateJavaVM(&jvm, &env, &vm_args);
#endif /* JNI_VERSION_1_2 */
  
  if(res == 0)
    staticJvm = jvm;
  return res;
}

/* Creates a Java VM and runs the static method FuseModule.init. If all goes well, the function
   returns 0. If the VM could not be started, it returns -2. If the class FuseModule could not be
   found, or the method can't be called it returns -1. */
static int init() {
  //FILE *debug = debugFile;
  DEBUG("  - Trying to get jvm\n");
  chdir(workingDirectory); // Working directory gets set to "/" in the new process. no good. I HOPE this won't affect the rest of the operation of fuse... I guess I'll notice.
  int cvmres = createJavaVM();
  if(cvmres < 0) {
    DEBUG("    - Could not create JVM! Exit code: %d\n", cvmres);
    //filler(buf, "ERROR_could_not_start_JVM", NULL, 0);
    return -2;
  }
  else {
    DEBUG("    - JVM created! Exit code: %d. Now trying to initialize FuseModule\n", cvmres);
    JNIEnv *env;
    (*staticJvm)->AttachCurrentThread(staticJvm, (void **)&env, NULL);
    printJVMInfo(staticJvm, env);
    DEBUG("    - Now trying to initialize FuseModule for real\n");
    jclass cls = (*env)->FindClass(env, "FuseModule");//org/apache/crimson/jaxp/SAXParserImpl");
    if ((*env)->ExceptionOccurred(env)) {
      (*env)->ExceptionDescribe(env);
      DEBUG("    - Exception occurred!\n");
    }
    
    if(cls != NULL) {
      DEBUG("    - Class FuseModule found\n");
      jmethodID mid = (*env)->GetStaticMethodID(env, cls, "init", "(Ljava/lang/String;)I");
      if(mid != NULL) {
	DEBUG("    - Method FuseModule.init found\n");
	jstring jpath = (*env)->NewStringUTF(env, rarFilename);
	if(jpath != NULL) {
	  DEBUG("    - String converted. Now calling CallStaticIntMethod\n");
	  jint status = (*env)->CallStaticIntMethod(env, cls, mid, jpath);
	  if(status == 0) {
	    DEBUG("    - Success!!!!!!\n");
	    return 0;
	  }
	  else if(status == -1) {
	    DEBUG("    - Could not open RAR... :(");
	  }
	}
      }
    }
    else {
      DEBUG("    - ERROR. Could not find class FuseModule\n");
    }
  }
  return -1;
}

static void mergeStat(JNIEnv *env, jobject statObject, struct stat *target) {
  DEBUG("    - Getting class Stat\n");
  jclass statClass = (*env)->FindClass(env, "Stat");
  DEBUG("      - statClass: %X\n", (int)statClass);
  DEBUG("    - Getting field id for st_mode\n");
  jfieldID modeFieldID = (*env)->GetFieldID(env, statClass, "st_mode", "S");
  DEBUG("    - Getting field id for st_nlink\n");
  jfieldID nlinkFieldID = (*env)->GetFieldID(env, statClass, "st_nlink", "B");
  DEBUG("    - Getting field id for st_size\n");
  jfieldID sizeFieldID = (*env)->GetFieldID(env, statClass, "st_size", "J");
  
  jshort modeValue = (*env)->GetShortField(env, statObject, modeFieldID);
  jbyte nlinkValue = (*env)->GetByteField(env, statObject, nlinkFieldID);
  jlong sizeValue = (*env)->GetLongField(env, statObject, sizeFieldID);

  target->st_mode = modeValue;
  target->st_nlink = nlinkValue;
  target->st_size = sizeValue;
  
  DEBUG("    - st_mode set to: %d\n", target->st_mode);
  DEBUG("    - st_nlink set to: %d\n", target->st_nlink);
  DEBUG("    - st_size set to: %d\n", (int)target->st_size); // Casted to int. fprintf doesn't seem to handle 64-bit integers

  // Now set time accordingly
  jfieldID cfid;
  cfid = (*env)->GetFieldID(env, statClass, "atime_tv_sec", "J"); target->st_atimespec.tv_sec = (*env)->GetLongField(env, statObject, cfid);
  cfid = (*env)->GetFieldID(env, statClass, "mtime_tv_sec", "J"); target->st_mtimespec.tv_sec = (*env)->GetLongField(env, statObject, cfid);
  cfid = (*env)->GetFieldID(env, statClass, "ctime_tv_sec", "J"); target->st_ctimespec.tv_sec = (*env)->GetLongField(env, statObject, cfid);
  target->st_atimespec.tv_nsec = 0;
  target->st_mtimespec.tv_nsec = 0;
  target->st_ctimespec.tv_nsec = 0;
  
  DEBUG(stderr, "    - st_atimespec.tv_sec set to: %d\n", target->st_atimespec.tv_sec);
  DEBUG(stderr, "    - st_mtimespec.tv_sec set to: %d\n", target->st_mtimespec.tv_sec);
  DEBUG(stderr, "    - st_ctimespec.tv_sec set to: %d\n", target->st_ctimespec.tv_sec);
}

int rarfs_getattr(const char *path, struct stat *stbuf)
{
  DEBUG("rarfs_getattr(\"%s\", ...);\n", path);
  int res = 0;
  
  memset(stbuf, 0, sizeof(struct stat));
  if(strcmp(path, "/") == 0) {
    stbuf->st_mode = S_IFDIR | 0755;
    stbuf->st_nlink = 2;
  }
  else if(HELLO_ENABLED && strcmp(path, hello_path) == 0) {
    stbuf->st_mode = S_IFREG | 0444;
    stbuf->st_nlink = 1;
    stbuf->st_size = strlen(hello_str);
  }
  else {
    // Query FuseModule for the attributes
    if(staticJvm == NULL && init() == -1) {
      DEBUG("- rarfs_getattr could not initialize jvm\n");
      staticJvm = NULL;
    }
    
    DEBUG("  - Trying to attach thread.\n");
    JNIEnv *env;
    (*staticJvm)->AttachCurrentThread(staticJvm, (void **)&env, NULL);
    DEBUG("  - Trying to get FuseModule class.\n");
    jclass cls = (*env)->FindClass(env, "FuseModule");
    if(env != NULL && cls != NULL) {
      DEBUG("  - env and cls aquired. Trying to get static method from env\n");
      jmethodID mid = (*env)->GetStaticMethodID(env, cls, "rarfsGetattr", "(Ljava/lang/String;LStat;)I");
      
      // Allocate new String containing the requested path
      DEBUG("  - Converting native string to Java String\n");
      jstring jpath = (*env)->NewStringUTF(env, path);
      DEBUG("    - jpath: %X", (int)jpath);

      // Allocate new Stat object
      DEBUG("  - Creating input Stat object\n");
      jclass statClass = (*env)->FindClass(env, "Stat");
      DEBUG("    - statClass: %X\n", (int)statClass);
      jmethodID statConstructor = (*env)->GetMethodID(env, statClass, "<init>", "()V");
      DEBUG("    - statConstructor: %X\n", (int)statConstructor);
      jobject statObject = (*env)->NewObject(env, statClass, statConstructor);
      DEBUG("    - statObject: %X\n", (int)statObject);
      
      if(mid != NULL && jpath != NULL && statObject != NULL) {
	DEBUG("  - Got prerequisites. Trying to call method \"public static Stat FuseModule.rarfsGetattr(String, Stat)\"...\n");
	jint returnValue = (*env)->CallStaticIntMethod(env, cls, mid, jpath, statObject);
	if(returnValue == 0) {
	  DEBUG("  - Method returned correctly, calling mergeStat\n");
	  mergeStat(env, statObject, stbuf);
	  DEBUG("    - mergeStat completed.\n");
	}
	else {
	  DEBUG("    - ERROR! returnValue=%d\n", (int)returnValue);
	  res = -ENOENT;
	}
      }
      else {
	DEBUG("    - ERROR! mid==%X jpath==%X statObject==%X\n", (int)mid, (int)jpath, (int)statObject);
	res = -ENOENT;
      }
      
      // Deallocate the allocated objects
      if(jpath != NULL)
	(*env)->DeleteLocalRef(env, jpath);
      if(statObject != NULL)
	(*env)->DeleteLocalRef(env, statObject);
    }
    else {
      DEBUG("    - ERROR! env==%X cls==%X", (int)env, (int)cls);
      res = -ENOENT;
    }
  }
  
  
  return res;
}

int rarfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
		  off_t offset, struct fuse_file_info *fi)
{
  /* Note 070221:
     What information do we need from the JVM? Two approaches:
     1. We just send everything on to the JVM, and handle everything there. The C code will
        merely act as a translator from C types to Java types.
     2. We simply request the information we need from the JVM and handle the information in
        C code.
     So which one is the best?
     If we do the data processing in Java, there's less probability of error, and data
     manipulation is easier (especially string operations). But how can we translate for
     instance the fuse_fill_dir_t, which is a function pointer, into Java...? That means we
     need JNI code for that as well.
     I still feel better about doing it that way....
     
     Note 070223:
     After some coding, I realize that I can't use the JVM used to start the program in the
     actual fuse module, as fuse terminates the startup process after calling fuse_main...
     thus also terminating the JVM. So the JVM will be started from the new process which has
     the purpose of serving the filesystem data.
  */
  
  DEBUG("rarfs_readdir\n");
/*   (void) offset; */
/*   (void) fi; */
  
/*   if(strcmp(path, "/") != 0) */
/*     return -ENOENT; */
  
  filler(buf, ".", NULL, 0);
  filler(buf, "..", NULL, 0);
  
  /* Now request the file list from the JVM. */
  if(staticJvm == NULL && init() == -1) {
    filler(buf, "could_not_initialize_jvm", NULL, 0);
    staticJvm = NULL;
  }
  else {
    DEBUG("  - Trying to attach thread (jvm: %X).\n", (int)staticJvm);
    JNIEnv *env;
    (*staticJvm)->AttachCurrentThread(staticJvm, (void **)&env, NULL);
    DEBUG("  - Trying to get FuseModule class.\n");
    jclass cls = (*env)->FindClass(env, "FuseModule");
    if(env != NULL && cls != NULL) {
      DEBUG("  - env and cls aquired. Trying to get static method from env\n");
      jmethodID mid = (*env)->GetStaticMethodID(env, cls, "rarfsReaddir", "(Ljava/lang/String;)[Ljava/lang/String;");
      
      // Allocate new String
      DEBUG("  - Converting native string to Java String\n");
      jstring jpath = (*env)->NewStringUTF(env, path);
      
      if(mid != NULL && jpath != NULL) {
	DEBUG("  - Got Java String. Trying to call static object method.\n");
	jobjectArray res = (*env)->CallStaticObjectMethod(env, cls, mid, jpath);
	DEBUG("  - Result from method call:\n");
	jint argc = (*env)->GetArrayLength(env, res);
	int i;
	for(i = 0; i < argc; ++i) {
	  jstring str = (*env)->GetObjectArrayElement(env, res, i);
	  const char *chars = (*env)->GetStringUTFChars(env, str, NULL);
	  DEBUG("    %d: %s\n", i, chars);
	  filler(buf, chars+1, NULL, 0);
	  (*env)->ReleaseStringUTFChars(env, str, chars);
	}
      }

      // Deallocate the allocated objects
      if(jpath != NULL)
	(*env)->DeleteLocalRef(env, jpath);
    }
  }
  
  if(HELLO_ENABLED)
    filler(buf, hello_path + 1, NULL, 0);
  
  return 0;
}

int rarfs_open(const char *path, struct fuse_file_info *fi)
{
  DEBUG("rarfs_open(\"%s\", ...);\n", path);
  if(!HELLO_ENABLED || strcmp(path, hello_path) != 0) {
    // Query FuseModule for the attributes
    if(staticJvm == NULL && init() == -1) {
      DEBUG("- rarfs_open could not initialize jvm!!! :( FUCK!\n");
      staticJvm = NULL;
    }
    
    int returnCode = 0;
    
    DEBUG("  - Trying to attach thread.\n");
    JNIEnv *env;
    (*staticJvm)->AttachCurrentThread(staticJvm, (void **)&env, NULL);
    DEBUG("  - Trying to get FuseModule class.\n");
    jclass cls = (*env)->FindClass(env, "FuseModule");
    if(env != NULL && cls != NULL) {
      DEBUG("  - env and cls aquired. Trying to get static method from env\n");
      jmethodID mid = (*env)->GetStaticMethodID(env, cls, "rarfsOpen", "(Ljava/lang/String;)I");
      DEBUG("  - Converting native string to Java String\n");
      jstring jpath = (*env)->NewStringUTF(env, path);
      if(mid != NULL && jpath != NULL) {
	DEBUG("  - Got Java String. Trying to call static open method.\n");
	jint exitcode = (*env)->CallStaticIntMethod(env, cls, mid, jpath);
	if(exitcode != 0)
	  returnCode = -ENOENT;
	else
	  DEBUG("  - Successfully opened!!\n");
      }
      else
	returnCode = -ENOENT;
      
      // Deallocate the allocated objects
      if(jpath != NULL)
	(*env)->DeleteLocalRef(env, jpath);
    }
    else
      returnCode = -ENOENT;
      
    return returnCode;
  }
  else if(HELLO_ENABLED) {
    if((fi->flags & 3) != O_RDONLY)
      return -EACCES;
    
    return 0;
  }
  else
    return -ENOENT;
}

int rarfs_read(const char *path, char *buf, size_t size, off_t offset,
	       struct fuse_file_info *fi)
{
  DEBUG("rarfs_read(\"%s\", ..., %d, %d, ...);\n", path, (int)size, (int)offset); // size and offset are 64-bit types I think. int-casted since printf won't deal with those (at least we can see something?).
  if(!HELLO_ENABLED || strcmp(path, hello_path) != 0) {
    if(staticJvm == NULL && init() == -1) {
      DEBUG("- rarfs_read could not initialize jvm!!! :( FUCK!\n");
      staticJvm = NULL;
    }
    int returnCode = -ENOENT;
    
    DEBUG("  - Trying to attach thread.\n");
    JNIEnv *env;
    (*staticJvm)->AttachCurrentThread(staticJvm, (void **)&env, NULL);
    DEBUG("  - Trying to get FuseModule class.\n");
    jclass cls = (*env)->FindClass(env, "FuseModule");
    if(env != NULL && cls != NULL) {
      DEBUG("  - Converting native string to Java String\n");
      jstring jpath = (*env)->NewStringUTF(env, path);
      if(USEDIRECTBUFFER) {
	DEBUG("  - Trying to get static method from env\n");
	jmethodID mid = (*env)->GetStaticMethodID(env, cls, "rarfsRead", "(Ljava/lang/String;Ljava/nio/ByteBuffer;JI)I");
	DEBUG("  - Mapping buf to a direct byte buffer\n");
	jobject directBuffer = (*env)->NewDirectByteBuffer(env, buf, size);
	if(mid != NULL && jpath != NULL && directBuffer != NULL) {
	  DEBUG("  - Trying to call static rarfsRead method.\n");
	  jint exitcode = (*env)->CallStaticIntMethod(env, cls, mid, jpath, directBuffer, offset, size);
	  if(exitcode < 0)
	    returnCode = -ENOENT;
	  else {
	    DEBUG("  - Read %d bytes into buf.\n", (int)exitcode); // I have no idea why jint needs to be casted. should be 32-bit signed.
	    returnCode = exitcode;
	  }
	}
	else
	  returnCode = -ENOENT;
	
	if(directBuffer != NULL)
	  (*env)->DeleteLocalRef(env, directBuffer);
      }
      else {
	DEBUG("  - Trying to get static method from env\n");
	jmethodID mid = (*env)->GetStaticMethodID(env, cls, "rarfsRead", "(Ljava/lang/String;[BJI)I");
	DEBUG("  - Creating new byte array.\n");
	jbyteArray cache = (*env)->NewByteArray(env, size);
	if(mid != NULL && jpath != NULL && cache != NULL) {
	  DEBUG("  - Got Java String. Trying to call static read method.\n");
	  jint exitcode = (*env)->CallStaticIntMethod(env, cls, mid, jpath, cache, offset, size);
	  if(exitcode < 0)
	    returnCode = -ENOENT;
	  else {
	    (*env)->GetByteArrayRegion(env, cache, 0, exitcode, (jbyte*)buf);
	    DEBUG("  - Read %d bytes into buf.\n", (int)exitcode); // I have no idea why jint needs to be casted. should be 32-bit signed.
	    returnCode = exitcode;
	  }
	}
	else
	  returnCode = -ENOENT;
	
	if(cache != NULL)
	  (*env)->DeleteLocalRef(env, cache);
      }
      
      // Deallocate the allocated objects
      if(jpath != NULL)
	(*env)->DeleteLocalRef(env, jpath);
    }
    else
      returnCode = -ENOENT;

    return returnCode;
  }
  else if(HELLO_ENABLED) {
    size_t len;
    //(void) fi; // what the _hell_ is this line here for?
    len = strlen(hello_str);
    if (offset < len) {
      if (offset + size > len)
	size = len - offset;
      memcpy(buf, hello_str + offset, size);
    } else
      size = 0;
    return size;
  }
  else return -ENOENT;
}

struct fuse_operations rarfs_oper = {
    .getattr	= rarfs_getattr,
    .readdir	= rarfs_readdir,
    .open	= rarfs_open,
    .read	= rarfs_read,
};

int rarfs_main(int argc, char *argv[])
{
  if(DEBUGMODE) {
    printf("argc: %d\n", argc);
    int i;
    for(i = 0; i < argc; ++i)
      printf("argv[%d]: %s\n", i, argv[i]);
  }
  
  rarFilename = argv[1];
  argv[1] = argv[0];
  return fuse_main(argc-1, argv+1, &rarfs_oper);
}

int main(int argc, char *argv[])
{
  if(getcwd(workingDirectory, MAXPATHLEN) == NULL) {
    fprintf(stderr, "Could not get working directory.");
    return -1;
  }
  else {
    if(DEBUGMODE) {
      fprintf(stderr, "Info: cwd = \"%s\"\n", workingDirectory);
      debugFile = stderr;//fopen("debug_log.txt", "a");
    }
    else
      debugFile = fopen("/dev/null", "a");
    
    return rarfs_main(argc, argv);
  }
}
